;; Generated package description from triples.el  -*- no-byte-compile: t -*-
(define-package "triples" "0.3" "A flexible triple-based database for use in apps" '((seq "2.0") (emacs "28.1")) :commit "45dc57703ae870aa864a195041b76d622535c351" :authors '(("Andrew Hyatt" . "ahyatt@gmail.com")) :maintainer '("Andrew Hyatt" . "ahyatt@gmail.com") :keywords '("triples" "kg" "data" "sqlite") :url "https://github.com/ahyatt/triples")
